# -*- coding: utf-8 -*-
"""
As BioSTEAM objects are created, they are automatically registered. The `find` object allows the user to find any Unit, Stream or System instance.  When `find` is called, it simply looks up the item and returns it. 

"""
from graphviz import Digraph
from IPython import display
from ._utils import Register, search_register

__all__ = ('find', 'Flowsheet')


def make_digraph(units, streams):
    """Return digraph of units and streams."""
    # Create a digraph and set direction left to right
    f = Digraph(format='svg')
    f.attr(rankdir='LR')
    # Set up unit nodes
    UD = {}  # Contains full description (ID and line) by ID
    for u in units:
        if hasattr(u, 'link_streams'): u._link_streams()
        graphics = u._graphics
        if not graphics.in_system:
            continue  # Ignore Unit
        
        # Initialize graphics and make Unit node with attributes
        Type = graphics.node_function(u) or u.line
        name = u.ID + '\n' + Type
        f.attr('node', **u._graphics.node)
        f.node(name)
        UD[u] = name
        
    keys = UD.keys()

    # Set attributes for graph and streams
    f.attr('node', shape='rarrow', fillcolor='#79dae8',
           style='filled', orientation='0', width='0.6',
           height='0.6', color='black', peripheries='1')
    f.attr('graph', splines='normal', overlap='orthoyx',
           outputorder='edgesfirst', nodesep='0.15', maxiter='1000000')
    f.attr('edge', dir='foward')
    
    for s in streams:
        if not s: continue  # Ignore stream

        oU = s._source
        if oU:
            oi = oU._outs.index(s) 
        
        dU = s._sink
        if dU:
            di = dU._ins.index(s)

        # Make stream nodes / unit-stream edges / unit-unit edges
        if oU not in keys and dU not in keys: pass
            # Stream is not attached to anything
        elif oU not in keys:
            # Feed stream case
            f.node(s.ID)
            edge_in = dU._graphics.edge_in
            f.attr('edge', arrowtail='none', arrowhead='none',
                   tailport='e', **edge_in[di])
            f.edge(s.ID, UD[dU])
        elif dU not in keys:
            # Product stream case
            f.node(s.ID)
            edge_out = oU._graphics.edge_out
            f.attr('edge', arrowtail='none', arrowhead='none',
                   headport='w', **edge_out[oi])
            f.edge(UD[oU], s.ID)
        else:
            # Process stream case
            edge_in = dU._graphics.edge_in
            edge_out = oU._graphics.edge_out
            f.attr('edge', arrowtail='none', arrowhead='normal',
                   **edge_in[di], **edge_out[oi])
            f.edge(UD[oU], UD[dU], label=s.ID)
    return f

def save_digraph(digraph, file, format):
    if not file:
        if format == 'svg':
            x = display.SVG(digraph.pipe(format=format))
        else:
            x = display.Image(digraph.pipe(format='png'))
        display.display(x)
    else:
        if '.' not in file:
            file += '.' + format
        img = digraph.pipe(format=format)
        f = open(file, 'wb')
        f.write(img)
        f.close()


# %% Flowsheet search      

class Flowsheets:
    __getattribute__ = __getitem__ = object.__getattribute__
    
    def __setattr__(self, key, value):
        raise TypeError(f"'{type(self).__name__}' object does not support attribute assignment")
    __setitem__ = __setattr__
    
    def __iter__(self):
        yield from self.__dict__.values()
    
    def __delattr__(self, key):
        if key == find.ID:
            raise AttributeError('cannot delete main flowsheet')
        else:
            super().__delattr__(key)
    
    def __repr__(self):
        return f'Register:\n ' + '\n '.join([repr(i) for i in self])
    
    
class Flowsheet:
    """Create a Flowsheet object which stores references to all stream, unit, and system objects."""
    
    #: [Register] All flowsheets.
    flowsheet = Flowsheets()
    
    def __init__(self, ID):        
        #: [Register] Contains all System objects as attributes.
        self.system = Register()
        
        #: [Register] Contains all Unit objects as attributes.
        self.unit = Register()
        
        #: [Register] Contains all Stream objects as attributes.
        self.stream = Register()
        
        #: [str] ID of flowsheet.
        self._ID = ID
        self.flowsheet.__dict__[ID] = self
    
    def __setattr__(self, key, value):
        if hasattr(self, '_ID'):
            raise TypeError(f"'{type(self).__name__}' object does not support attribute assignment")
        else:
            super().__setattr__(key, value)
    
    @property
    def ID(self):
        return self._ID
    
    def diagram(self, kind='surface'):
        """Display all units and attached streams.
        
        **kind:** Must be one of the following:
                * **'thorough':** Thoroughly display every unit.
                * **'surface':** Display units and recycle systems.
                * **'minimal':** Minimally display flowsheet as a box with feeds and products.
        
        """
        if kind == 'thorough':
            return self._thorough_diagram()
        elif kind == 'surface':
            return self._surface_diagram()
        elif kind == 'minimal':
            return self._minimal_diagram()
        else:
            raise ValueError(f"kind must be either 'thorough', 'surface', or 'minimal'.")
    
    def _thorough_diagram(self, file=None, format='svg'):
        units = list(self.unit)
        units.reverse()
        streams = set()
        for u in units:
            streams.update(u._ins)
            streams.update(u._outs)
        f = make_digraph(units, streams)
        save_digraph(f, file, format)
    
    def _minimal_diagram(self):
        from . import _system, Stream
        streams = list(self.stream)
        feeds = set(filter(_system._isfeed, streams))
        products = set(filter(_system._isproduct, streams))
        product = Stream(None)
        product._ID = ''
        feed = Stream(None)
        feed._ID = ''
        _system._streamUnit('\n'.join([i.ID for i in feeds]),
                           None, feed)
        _system._streamUnit('\n'.join([i.ID for i in products]),
                           product, None)
        unit = _system._systemUnit(self.ID, feed, product)
        unit.line = 'flowsheet'
        unit.diagram(1)
        
    def _surface_diagram(self, file=None, format='svg'):
        from . import _system, Stream
        units = set(self.unit)
        StrUnit = _system._streamUnit
        refresh_units = set()
        for i in self.system:
            if i.recycle and not any(sub.recycle for sub in i.subsystems):
                outs = []
                ins = []
                feeds = []
                products = []
                for s in i.streams:
                    source = s._source
                    sink = s._sink
                    if source in i.units and sink not in i.units:
                        if sink: outs.append(s)
                        else: products.append(s)
                        refresh_units.add(source)
                    elif sink in i.units and source not in i.units:
                        if source: ins.append(s)
                        else: feeds.append(s)
                        refresh_units.add(sink)
                
                if len(feeds) > 1:
                    feed = Stream(None)
                    feed._ID = ''
                    units.add(StrUnit('\n'.join([i.ID for i in feeds]),
                                      None, feed))
                    ins.append(feed)
                else: ins += feeds
                
                if len(products) > 1:
                    product = Stream(None)
                    product._ID = ''
                    units.add(StrUnit('\n'.join([i.ID for i in products]),
                                      product, None))
                    outs.append(product)
                else: outs += products
                
                subsystem_unit = _system._systemUnit(i.ID, ins, outs)
                units.difference_update(i.units)
                units.add(subsystem_unit)
        
        sys = _system.System(None, units)
        sys._thorough_diagram(file, format)
        for i in refresh_units:
            i._ins[:] = i._ins
            i._outs[:] = i._outs
    
    def __call__(self, ID):
        """Return requested biosteam item.
    
        **Parameters**
    
             **ID:** [str] ID of the requested item.
    
        """
        ID = ID.replace(' ', '_')
        obj = (search_register(self.stream, ID)
               or search_register(self.unit, ID)
               or search_register(self.system, ID))
        if not obj: raise LookupError(f"no registered item '{ID}'")
        return obj
    
    def __str__(self):
        return self.ID
    
    def __repr__(self):
        return f'<{type(self).__name__}: {self.ID}>'


class MainFlowsheet(Flowsheet):
    """Create a MainFlowsheet object which automatically registers biosteam objects as they are created."""
    __slots__ = ()
    
    @staticmethod
    def set_flowsheet(flowsheet):
        """Set main flowsheet that is updated with new biosteam objects."""
        if isinstance(flowsheet, Flowsheet):
            object.__setattr__(find, '__dict__', flowsheet.__dict__)
        else:
            raise TypeError('flowsheet must be a Flowsheet object')
    
    __setattr__ = Flowsheets.__setattr__
    
    def __new__(cls):
        raise TypeError('cannot create new find object.')

    def __repr__(self):
        return f'<{type(self).__name__}: {self.ID}>'
    
    
#: [find] Find BioSTEAM objects by ID.
find = object.__new__(MainFlowsheet)
find.set_flowsheet(Flowsheet('Default'))
    

# %% Attempt at contantly rendered digraph

# #: [bool] True if flowsheet is contantly rendered
# self.live = True

# def _show_once(self):
    #     fig = plt.figure()
    #     plt.axis('off')
    #     f = open('diagram.png', 'wb')
    #     diagram = make_digraph(self.unit.values()).pipe(format='png')
    #     f.write(diagram)
    #     img = mpimg.imread('diagram.png')
    #     plt.imshow(img, interpolation='spline36')
    #     fig.show()
    #     # img = Image.open('diagram.png')
    #     # img.show() 
    
    # def _show(self):
    #     fig = plt.figure()
    #     plt.axis('off')
    #     f = open('diagram.png', 'wb')
    #     for i in range(3):
    #         diagram = make_digraph(self.unit.values()).pipe(format='png')
    #         f.write(diagram)
    #         img = mpimg.imread('diagram.png')
    #         plt.clf() 
    #         plt.imshow(img, interpolation='spline36')
    #         fig.show()
    #         time.sleep(10)
    #     f.close()
    #     os.remove('diagram.png')
    
    # def show(self):
    #     t1 = threading.Thread(target=self._show())
    #     t1.start()


# -*- coding: utf-8 -*-
"""
Created on Sat Nov 10 19:27:01 2018

@author: yoelr
"""
from ._compound import Compound, _R
from ._chemical import Chemical
import numpy as np
from math import log

__all__ = ('Substance', 'StaticChemical')


def select_value(obj, user_defined, attr, default):
    if user_defined is not None: return user_defined
    try:
        obj_value = getattr(obj, attr)
        if obj_value is None:
            return default
        else:
            return obj_value
    except:
        return default
    
# %%

class Substance(Compound):
    """Create a Substance object with the same material properties as "obj" argument for simplified modeling purposes. The phase stays constant as the specified `default_phase`.

    **Parameters**
    
        **ID:** [str] ID of specie
        
        **CAS:** [str] CAS identification
        
        **obj** [Specie] Should contain analog properties for specie at STP. Defaults to water if None.

        **MW:** Molecular weight of compound (defaults to 1)
         
        **T:** Temperature (K)
         
        **P:** Pressure (Pa)
         
        **rho:** Density (kg/m^3)
        
        **k:** Thermal conductivity (W/m/K)
        
        **mu:** Hydrolic viscosity (Pa*s)
        
        **sigma:** Surface tension (N/m)
        
        **Hfm:** Heat of formation (J/mol)
        
        **Hc:** Heat of combustion (J/mol)
        
        **phase:** {'l', 's', 'g'} Phase of substance
        
        ****properties:** Additional properties to set
        
    **Examples**
    
        Create a 'Yeast' substance with the same properties as water:
    
        .. code-block:: python
            
            >>> from biosteam.compounds import Substance
            >>> Yeast = Substance('Yeast', obj=Chemical('Water'))
            >>> Yeast.rho
            997
            
    """
    def __init__(self, ID, CAS='', obj=None, *,
                 MW=1, T=298.15, P=101325, Cp=None,
                 rho=None, k=None, mu=None, sigma=None,
                 Pr=None, Hf=None, Hc=None, phase=None, **properties):
        if obj:
            obj.T = T
            obj.P = P
            if phase:
                obj.phase = phase
            else:
                phase = obj.phase
        if not phase: phase = 'l'
        
        self.ID = ID.replace(' ', '_')
        self.T = T
        self.P = P
        self.Hvapm = 0
        
        # Set property values (defaults to Water property values)
        self.MW = MW = select_value(obj, MW, 'MW', 1)
        self.CAS = select_value(obj, CAS, 'CAS', ID)
        self.Cp = select_value(obj, Cp, 'Cp', 4.18)
        self.sigma = select_value(obj, sigma, 'mu', 0.072055)
        self.rho = select_value(obj, rho, 'rho', 997)
        self.mu = select_value(obj, mu, 'mu', 0.00091272)
        self.k = select_value(obj, k, 'k', 0.5942)
        self.Hf = select_value(obj, Hf, 'Hf', 0)
        self.Hc = select_value(obj, Hc, 'Hc', 0)
        self.Hfm = self.Hf*MW
        self.Hcm = self.Hc*MW
        self.rhom = self.rho/MW*1000
        self.Vm = 1/self.rhom
        self.Cpm = self.Cp * MW
        self.alpha = self.k/self.rho/self.Cp
        self.nu = self.mu/self.rho
        self.Pr = self.nu/self.alpha
        self.UNIFAC_groups = self.UNIFAC_Dortmund_groups = None
        self.phase_ref = phase
        
        if phase not in 'slg':
            raise ValueError(f"default phase must be either 's', 'l' or 'g', not '{phase}'")
        
        # Phase change temperature consitent with phase
        self.Tb = 0 if phase=='g' else np.inf 
        self.Tm = np.inf if phase=='s' else 0
        for i in properties: setattr(self, i, properties[i])
        
    @property
    def H(self):
        return self.Cpm*(self.T-self.T_ref)
    @property
    def S(self):
        return self.Cpm*log(self.T/self.T_ref) + (_R*log(self.P/self.P_ref) if self.phase=='g' else 0)
    
    @property
    def phase(self):
        """Phase of substance."""
        return self.phase_ref
    @phase.setter
    def phase(self, phase): pass


class StaticChemical(Substance):
    def __init__(self, ID, *, T=298.15, P=101325, MW=None,
                 Cp=None, rho=None, k=None, mu=None, sigma=None,
                 Pr=None, Hf=None, Hc=None, phase=None, **properties):
        obj = Chemical(ID, T, P)
        super().__init__(obj=obj,
                 ID=obj.ID, CAS=obj.CAS, MW=obj.MW, T=T, P=P,
                 Cp=Cp, rho=rho, k=k, mu=mu, sigma=sigma,
                 Pr=Pr, Hf=Hf, Hc=Hc, phase=phase, **properties)


# -*- coding: utf-8 -*-
"""
Created on Mon Apr 29 18:28:55 2019

@author: yoelr
"""
from . import Facility
from ..decorators import cost

__all__ = ('ChilledWaterPackage',)

@cost('Duty', 'Cooling tower',
      S=-14*4184000, kW=3400*0.7457, cost=1375e3, CE=551, n=0.7, BM=1.5)
class ChilledWaterPackage(Facility):
    """Create a chilled water package that is cost based on flow rate of chilled water.
    
    **Parameters**
    
        **Duty:** Chilled water duty (kJ/hr)
    
    """
    _N_heat_utilities = 1
    _N_ins = _N_outs = 0
    _units = {'Duty': 'kJ/h4'}
    def __init__(self, ID=''):
        Facility.__init__(self, ID, None, None)
        self.chilled_water_utilities = set()
        
    def _design(self):
        cwu = self.chilled_water_utilities
        if not cwu:
            for u in self.system.units:
                if u is self: continue
                for hu in u._heat_utilities:
                    if hu.ID == 'Chilled water':
                        cwu.add(hu)
        self._Design['Duty'] = duty = sum(i.duty for i in cwu)
        self._heat_utilities[0](duty, 330)
        
        


# -*- coding: utf-8 -*-
"""
Created on Tue Jul  9 00:35:01 2019

@author: yoelr
"""
from .._exceptions import SolverError
#from .other_utils import count
import numpy as np

__all__ = ('bounded_secant', 'secant', 'wegstein', 'conditional_wegstein',
           'accelerated_secant', 'accelerated_bounded_secant')

def secant(f, x0, x1, xtol, ytol=1e-7, args=(), maxiter=50):
    pos = abs
    y0 = f(x0, *args)
    if pos(y0) < ytol: return x0
    y1 = f(x1, *args) 
    if pos(y1) < ytol: return x1
    x_diff = x1-x0 
    for iter in range(maxiter): 
        x1 = x0 - y1*(x_diff)/(y1-y0)
        x_diff = x1-x0 
        if pos(x_diff) < xtol or pos(y1) < ytol: return x1
        x0 = x1
        y0 = y1
        y1 = f(x1, *args)
    raise SolverError(f'failed to converge after {maxiter} iterations')

def bounded_secant(f, x0, x, x1, xtol, ytol=1e-7, args=()):
    y0 = f(x0, *args)
    y1 = f(x1, *args)
    assert y0*y1 < 0, ('objective function bounderies '
                       'must have opposite signs')
    if y1 < 0: x0, y0, x1, y1 = x1, y1, x0, y0
    x_ = x - 2*xtol
    while abs(x-x_) > xtol:
        y = f(x, *args)
        if y > ytol:
            x_ = x1 = x
            y1 = y
        elif y < ytol:
            x_ = x0 = x
            y0 = y
        else: return x
        x = x0 - y0*(x1-x0)/(y1-y0)
    return x

def accelerated_secant(f, x0, x1, xtol, ytol=1e-7, args=(), maxiter=50):
    pos = abs
    y0 = f(x0, *args)
    if pos(y0) < ytol: return x0
    y1 = f(x1, *args)
    if pos(y1) < ytol: return x1 
    x1 = g0 = x0 - y1*(x1-x0)/(y1-y0)
    if (pos(x1-x0) < xtol): return x1
    y0 = y1
    for iter in range(maxiter): 
        y1 = f(x1, *args)
        g1 = x1 - y1*(x1-x0)/(y1-y0)
        if (pos(g1-x1) < xtol) or (pos(y1) < ytol): return g1
        w = (x1-x0)/(x1-g1 + g0-x0)
        x1 = w*g1 + (1-w)*x1
        x0 = x1
        y0 = y1
        g0 = g1
    raise SolverError(f'failed to converge after {maxiter} iterations')

def accelerated_bounded_secant(f, x0, x, x1, xtol, ytol=1e-7, args=(), maxiter=50):
    y0 = f(x0, *args)
    y1 = f(x1, *args)
    assert y0*y1 < 0, ('objective function bounderies '
                       'must have opposite signs')
    if y1 < 0: x0, y0, x1, y1 = x1, y1, x0, y0
    
    x_old = x
    y = f(x, *args)
    if y > ytol:
        x1 = x
        y1 = y
    elif y < ytol:
        x0 = x
        y0 = y
    else: return x
    x = g0 = x0 - y0*(x1-x0)/(y1-y0)
    
    while abs(x-x_old) > xtol:
        y = f(x, *args)
        if y > ytol:
            x1 = x
            y1 = y
        elif y < ytol:
            x0 = x
            y0 = y
        else: return x
        g1 = x0 - y0*(x1-x0)/(y1-y0)
        w = (x-x_old)/(x-g1 + g0-x_old)
        x_old = x
        x = w*g1 + (1-w)*x
        if x < x0 or x > x1: x = g1
        g0 = g1
    return x

def wegstein(f, x0, xtol, args=(), maxiter=50):
    x1 = y0 = f(x0, *args)
    for iter in range(maxiter):
        if (abs(y0 - x0) < xtol).all(): return x1
        try: y1 = f(x1, *args)
        except:
            x1 = y1
            y1 = f(x1, *args)
        w = (x1-x0)/(x1-y1 + y0-x0)
        x0 = x1
        y0 = y1
        w[np.logical_not(np.isfinite(w))] = 1
        x1 = w*y1 + (1-w)*x1
    raise SolverError(f'failed to converge after {maxiter} iterations')

def conditional_wegstein(f, x0):
    y0, condition = f(x0)
    x1 = y0
    while condition:
        try: y1, condition = f(x1)
        except:
            x1 = y1
            y1, condition = f(x1)
        w = (x1-x0)/(x1-y1 + y0-x0)
        x0 = x1
        y0 = y1
        w[np.logical_not(np.isfinite(w))] = 1
        x1 = w*y1 + (1-w)*x1
    return x1


    

#    if y0*y1 < 0:
#             x = x0 - y1*(x1-x0)/(y1-y0)
#             return bounded_secant(f, x0, x, x1, y0, y1, xtol, ytol, args, maxiter-iter)


